@@ -27,7 +27,7 @@ static char* help_msg =
 "This program is used to enhance the automated testing.  It is not useful for\n"
 "anything much else.\n";
 
-#define CLEAR_BUFFER_SIZE		(2 * 1024 * 1024)
+#define CLEAR_BUFFER_SIZE		(1024 * 1024)
 #define CLEAR_BUFFER_SECTORS		(CLEAR_BUFFER_SIZE/512)
 
 static char buffer [CLEAR_BUFFER_SIZE];
@@ -173,7 +173,7 @@ _clear_unused_clusters (PedFileSystem* fs)
 {
 	FatSpecific*	fs_info = FAT_SPECIFIC (fs);
 	FatCluster	cluster;
-	FatCluster	run_start;
+	FatCluster	run_start = 0; /* shut gcc up! */
 	FatCluster	run_length = 0;
 
 	for (cluster = 2; cluster < fs_info->cluster_count + 2; cluster++) {
