@@ -61,8 +61,8 @@ typedef enum {
 	PED_DISK_TYPE_EXTENDED=1,
 	PED_DISK_TYPE_PARTITION_NAME=2
 } PedDiskTypeFeature;
-#define PED_DISK_TYPE_FIRST_FEATURE	PED_PARTITION_EXTENDED
-#define PED_DISK_TYPE_LAST_FEATURE	PED_PARTITION_NAME
+#define PED_DISK_TYPE_FIRST_FEATURE	PED_DISK_TYPE_EXTENDED
+#define PED_DISK_TYPE_LAST_FEATURE	PED_DISK_TYPE_PARTITION_NAME
 
 struct _PedPartition {
 	PedPartition*		prev;
@@ -152,7 +152,11 @@ extern int ped_disk_close (PedDisk* disk);
 extern int ped_disk_read (PedDisk* disk);
 extern int ped_disk_write (PedDisk* disk);
 extern int ped_disk_is_busy (const PedDisk* disk);
+
+int ped_disk_get_primary_partition_count (PedDisk* disk);
+int ped_disk_get_last_partition_num (PedDisk* disk);
 extern int ped_disk_get_max_primary_partition_count (const PedDisk* disk);
+
 extern PedDisk* ped_disk_alloc (PedDevice* dev, PedDiskType* type);
 extern void ped_disk_free (PedDisk* disk);
 
