@@ -1,6 +1,6 @@
 /*
     libparted - a library for manipulating disk partitions
-    Copyright (C) 1998-2000 Free Software Foundation, Inc.
+    Copyright (C) 2000 Free Software Foundation, Inc.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -32,5 +32,8 @@ typedef struct {
 
 #define PC98_NAME	"pc98"
 
+extern void ped_disk_pc98_init ();
+extern void ped_disk_pc98_done ();
+
 #endif /* PED_DISK_DOS_H_INCLUDED */
 
