@@ -32,7 +32,7 @@
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <string.h>
+#include <limits.h>
 #include <string.h>
 #include <unistd.h>
 #include <sys/ioctl.h>
@@ -54,6 +54,14 @@
 #define BLKGETSIZE _IO(0x12,96)	/* return device size */
 #define BLKFLSBUF  _IO(0x12,97)	/* flush buffer cache */
 #define BLKSSZGET  _IO(0x12,104) /* get block device sector size */
+#define BLKGETLASTSECT  _IO(0x12,108) /* get last sector of block device */
+#define BLKSETLASTSECT  _IO(0x12,109) /* set last sector of block device */
+
+struct blkdev_ioctl_param {
+	unsigned int block;
+	size_t content_length;
+	char * block_contents;
+};
 
 /* from <linux/major.h> */
 #define IDE0_MAJOR		3
@@ -79,7 +87,24 @@
 #define COMPAQ_SMART2_MAJOR5	77
 #define COMPAQ_SMART2_MAJOR6	78
 #define COMPAQ_SMART2_MAJOR7	79
+#define COMPAQ_SMART_MAJOR	104
+#define COMPAQ_SMART_MAJOR1	105
+#define COMPAQ_SMART_MAJOR2	106
+#define COMPAQ_SMART_MAJOR3	107
+#define COMPAQ_SMART_MAJOR4	108
+#define COMPAQ_SMART_MAJOR5	109
+#define COMPAQ_SMART_MAJOR6	110
+#define COMPAQ_SMART_MAJOR7	111
 #define DAC960_MAJOR		48
+#define ATARAID_MAJOR		114
+#define I2O_MAJOR1		80
+#define I2O_MAJOR2		81
+#define I2O_MAJOR3		82
+#define I2O_MAJOR4		83
+#define I2O_MAJOR5		84
+#define I2O_MAJOR6		85
+#define I2O_MAJOR7		86
+#define I2O_MAJOR8		87
 
 #define SCSI_BLK_MAJOR(M) (						\
 		(M) == SCSI_DISK0_MAJOR 				\
@@ -131,6 +156,102 @@ ped_device_get_next (const PedDevice* dev)
 		return devices;
 }
 
+/* wrap around POSIX brain-damage */
+static int
+_readlink (const char* path, char* buf, size_t size)
+{
+	memset (buf, 0, size);
+	return readlink (path, buf, size);
+}
+
+static void
+_strcut (char* str, size_t count)
+{
+	size_t	i;
+
+	for (i = count; str [i]; i++)
+		str [i - count] = str [i];
+	str [i - count] = 0;
+}
+
+static void
+_remove_double_slash (char* path)
+{
+	char*	match;
+
+	while ((match = strstr (path, "//")))
+		_strcut (match, 1);
+}
+
+/* removes ./ and XX/../ from path.  Obviously broken (will remove a./),
+ * but who cares?  Expects // to have been removed
+ */
+static void
+_remove_dots (char* path)
+{
+	char*	match;
+
+	match = strchr (path, '.');
+	while (match) {
+		if (!strncmp (match, "./", 2)) {
+			_strcut (match, 2);
+		} else if (!strncmp (match, "../", 3)) {
+			char*	cut_start;
+
+			for (cut_start = match - 2;
+			     *cut_start != '/'; cut_start--) {
+				if (cut_start < path)
+					return;	/* broken path! */
+			}
+
+			_strcut (cut_start, match - cut_start + 2);
+		}
+
+		match = strchr (path, '.');
+	}
+}
+
+/* follows symbolic links.  WTF isn't this in libc?! */
+static char*
+_normalize_path (const char* path)
+{
+	char	normalized [_POSIX_PATH_MAX + 1];
+	int	normalized_len;
+	char	tmp [_POSIX_PATH_MAX + 1];
+
+	strcpy (normalized, path);
+	_remove_double_slash (normalized);
+	_remove_dots (normalized);
+	normalized_len = strlen (path);
+
+	while (_readlink (normalized, tmp, _POSIX_PATH_MAX + 1) > 0) {
+		if (tmp [0] == '/') {
+			strcpy (normalized, tmp);
+		} else {
+			char* slash_match = strrchr (normalized, '/');
+			char* cpy_loc;
+
+			if (slash_match) {
+				if (slash_match - normalized + normalized_len
+						> _POSIX_PATH_MAX)
+					return NULL;
+				cpy_loc = slash_match + 1;
+			} else {
+				cpy_loc = normalized;
+			}
+
+			/* check for recursive sym-links */
+			if (!strcmp (cpy_loc, tmp))
+				return NULL;
+			strcpy (cpy_loc, tmp);
+		}
+		_remove_double_slash (normalized);
+		_remove_dots (normalized);
+		normalized_len = strlen (path);
+	}
+	return strdup (normalized);
+}
+
 /* First searches through probed devices, then attempts to open the device
  * regardless.
  */
@@ -138,13 +259,20 @@ PedDevice*
 ped_device_get (const char* path)
 {
 	PedDevice*	walk;
+	char*		normalized;
 
 	PED_ASSERT (path != NULL, return NULL);
 
+	normalized = _normalize_path (path);
+	if (!normalized)
+		return NULL;
 	for (walk = devices; walk != NULL; walk = walk->next) {
-		if (!strcmp (walk->path, path))
+		if (!strcmp (walk->path, normalized)) {
+			ped_free (normalized);
 			return walk;
+		}
 	}
+	ped_free (normalized);
 
 	walk = ped_device_new (path);
 	if (!walk)
@@ -213,7 +341,15 @@ _is_ide_major (int major)
 static int
 _is_cpqarray_major (int major)
 {
-	return COMPAQ_SMART2_MAJOR <= major && major <= COMPAQ_SMART2_MAJOR7;
+	return ((COMPAQ_SMART2_MAJOR <= major && major <= COMPAQ_SMART2_MAJOR7)
+	     || (COMPAQ_SMART_MAJOR <= major && major <= COMPAQ_SMART_MAJOR7));
+}
+
+
+static int
+_is_i2o_major (int major)
+{
+  	return (I2O_MAJOR1 <= major && major <= I2O_MAJOR8);
 }
 
 static int
@@ -241,6 +377,10 @@ device_probe_type (PedDevice* dev)
 		dev->type = PED_DEVICE_IDE;
 	} else if (dev_major == DAC960_MAJOR && (dev_minor % 0x8 == 0)) {
 		dev->type = PED_DEVICE_DAC960;
+	} else if (dev_major == ATARAID_MAJOR && (dev_minor % 0x10 == 0)) {
+		dev->type = PED_DEVICE_ATARAID;
+	} else if (_is_i2o_major (dev_major) && (dev_minor % 0x10 == 0)) {
+		dev->type = PED_DEVICE_I2O;
 	} else if (_is_cpqarray_major (dev_major) && (dev_minor % 0x10 == 0)) {
 		dev->type = PED_DEVICE_CPQARRAY;
 	} else {
@@ -529,6 +669,7 @@ init_generic (PedDevice* dev, char* model_name)
 			case PED_EXCEPTION_UNHANDLED:
 				ped_exception_catch ();
 			case PED_EXCEPTION_IGNORE:
+			        ; // just workaround for gcc 3.0
 		}
 
 		/* what should we stick in here? */
@@ -562,7 +703,10 @@ ped_device_new (const char* path)
 	if (!dev)
 		goto error;
 
-	dev->path = strdup (path);
+	dev->path = _normalize_path (path);
+	if (!dev->path)
+		goto error_free_dev;
+
 	dev->open_count = 0;
 	dev->external_mode = 0;
 	dev->dirty = 0;
@@ -571,32 +715,42 @@ ped_device_new (const char* path)
 	dev->geom_already_guessed = 0;
 
 	if (!device_probe_type (dev))
-		goto error_free_dev;
+		goto error_free_path;
 
 	switch (dev->type) {
 	case PED_DEVICE_IDE:
 		if (!init_ide (dev))
-			goto error_free_dev;
+			goto error_free_path;
 		break;
 
 	case PED_DEVICE_SCSI:
 		if (!init_scsi (dev))
-			goto error_free_dev;
+			goto error_free_path;
 		break;
 
 	case PED_DEVICE_DAC960:
 		if (!init_generic (dev, _("DAC960 RAID controller")))
-			goto error_free_dev;
+			goto error_free_path;
 		break;
 
 	case PED_DEVICE_CPQARRAY:
 		if (!init_generic (dev, _("Compaq Smart Array")))
-			goto error_free_dev;
+			goto error_free_path;
 		break;
 
+	case PED_DEVICE_ATARAID:
+		if (!init_generic (dev, _("ATARAID Controller")))
+			goto error_free_path;
+		break;
+
+	case PED_DEVICE_I2O:
+		if (!init_generic (dev, _("I2O Controller")))
+			goto error_free_dev;
+		break;
+		
 	case PED_DEVICE_FILE:
 		if (!init_file (dev))
-			goto error_free_dev;
+			goto error_free_path;
 		break;
 
 	case PED_DEVICE_UNKNOWN:
@@ -606,19 +760,21 @@ ped_device_new (const char* path)
 			_("Device %s is neither a SCSI nor IDE drive."),
 			dev->path)
 				== PED_EXCEPTION_CANCEL)
-			goto error_free_dev;
+			goto error_free_path;
 		if (!init_generic (dev, _("Unknown")))
-			goto error_free_dev;
+			goto error_free_path;
 		break;
 
 	default:
 		ped_exception_throw (PED_EXCEPTION_NO_FEATURE,
 				PED_EXCEPTION_CANCEL,
 				_("ped_device_new()  Unsupported device type"));
-		goto error_free_dev;
+		goto error_free_path;
 	}
 	return dev;
 
+error_free_path:
+	ped_free (dev->path);
 error_free_dev:
 	ped_free (dev);
 error:
@@ -637,33 +793,53 @@ ped_device_destroy (PedDevice* dev)
 	ped_free (dev);
 }
 
+char*
+ped_device_get_part_path (PedDevice* dev, int num)
+{
+	int		result_len = strlen (dev->path) + 16;
+	char*		result;
+
+	result = (char*) ped_malloc (result_len);
+	if (!result)
+		return NULL;
+
+	if (strstr (dev->path, "disc"))		/* HACK: devfs */
+	        snprintf (result, result_len, "%s/part%d", dev->path, num);
+	else if (dev->type == PED_DEVICE_DAC960
+			|| dev->type == PED_DEVICE_CPQARRAY
+			|| dev->type == PED_DEVICE_ATARAID)
+	        snprintf (result, result_len, "%sp%d", dev->path, num);
+	else
+	        snprintf (result, result_len, "%s%d", dev->path, num);
+
+	return result;
+}
+
 #ifdef linux
 static void
 _flush_cache (PedDevice* dev)
 {
 	int		i;
-	char*		name;
-	int		fd;
 
 	if (dev->read_only)
 		return;
 
 	ioctl (dev->fd, BLKFLSBUF);
 
-	name = (char*) ped_malloc (strlen (dev->path) + 3);
-	if (!name)
-		return;
-
 	for (i = 0; i < 16; i++) {
-		sprintf (name, "%s%d", dev->path, i);
+		char*		name;
+		int		fd;
+
+		name = ped_device_get_part_path (dev, i);
+		if (!name)
+			break;
 		fd = open (name, O_WRONLY, 0);
+		ped_free (name);
 		if (fd == -1)
 			continue;
 		ioctl (fd, BLKFLSBUF);
 		close (fd);
 	}
-
-	ped_free (name);
 }
 
 static int
@@ -871,6 +1047,69 @@ ped_device_seek (const PedDevice* dev, PedSector sector)
 	}
 }
 
+#ifdef linux
+static int
+_linux_read_lastoddsector (const PedDevice* dev, void* buffer)
+{
+	struct blkdev_ioctl_param ioctl_param;
+
+	PED_ASSERT(dev != NULL, return 0);
+	PED_ASSERT(buffer != NULL, return 0);
+
+retry:
+	ioctl_param.block = 0; /* read the last sector */
+	ioctl_param.content_length = dev->sector_size;
+	ioctl_param.block_contents = buffer;
+	
+	if (ioctl(dev->fd, BLKGETLASTSECT, &ioctl_param) == -1) {
+		PedExceptionOption	opt;
+		opt = ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_RETRY_IGNORE_CANCEL,
+			_("%s during read on %s"),
+			strerror (errno), dev->path);
+
+		if (opt == PED_EXCEPTION_CANCEL)
+			return 0;
+		if (opt == PED_EXCEPTION_RETRY)
+			goto retry;
+	}
+
+	return 1;
+}
+
+static int
+_linux_write_lastoddsector (const PedDevice* dev, const void* buffer)
+{
+	struct blkdev_ioctl_param ioctl_param;
+
+	PED_ASSERT(dev != NULL, return 0);
+	PED_ASSERT(buffer != NULL, return 0);
+
+retry:
+	ioctl_param.block = 0; /* write the last sector */
+	ioctl_param.content_length = dev->sector_size;
+	ioctl_param.block_contents = (void*) buffer;
+	
+	if (ioctl(dev->fd, BLKSETLASTSECT, &ioctl_param) == -1) {
+		PedExceptionOption	opt;
+		opt = ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_RETRY_IGNORE_CANCEL,
+			_("%s during write on %s"),
+			strerror (errno), dev->path);
+
+		if (opt == PED_EXCEPTION_CANCEL)
+			return 0;
+		if (opt == PED_EXCEPTION_RETRY)
+			goto retry;
+	}
+
+	return 1;
+}
+
+#endif /* linux */
+
 int
 ped_device_read (const PedDevice* dev, void* buffer, PedSector start,
 		 PedSector count)
@@ -883,6 +1122,20 @@ ped_device_read (const PedDevice* dev, void* buffer, PedSector start,
 	PED_ASSERT (!dev->external_mode, return 0);
 	PED_ASSERT (buffer != NULL, return 0);
 
+#ifdef linux
+	/* Kludge.  This is necessary to read/write the last
+	   block of an odd-sized disk, until Linux 2.5.x kernel fixes.
+	*/
+
+	if (dev->type != PED_DEVICE_FILE
+	    && (dev->length & 1)
+	    && start + count - 1 == dev->length - 1) {
+		return ped_device_read (dev, buffer, start, count - 1)
+			&& _linux_read_lastoddsector (dev,
+					buffer + (count-1) * 512);
+	}
+#endif
+
 	while (1) {
 		if (ped_device_seek (dev, start))
 			break;
@@ -1003,6 +1256,20 @@ ped_device_write (PedDevice* dev, const void* buffer, PedSector start,
 			return 1;
 	}
 
+#ifdef linux
+	/* Kludge.  This is necessary to read/write the last
+	   block of an odd-sized disk, until Linux 2.5.x kernel fixes.
+	*/
+
+	if (dev->type != PED_DEVICE_FILE
+	    && (dev->length & 1)
+	    && start + count - 1 == dev->length - 1) {
+		return ped_device_write (dev, buffer, start, count - 1)
+			&& _linux_write_lastoddsector (dev,
+					buffer + (count-1) * 512);
+	}
+#endif
+
 	while (1) {
 		if (ped_device_seek (dev, start))
 			break;
@@ -1095,34 +1362,34 @@ ped_device_check (PedDevice* dev, void* buffer, PedSector start,
 }
 
 static void
-probe (char* path)
+probe (const char* path)
 {
 	PedDevice*	dev;
+	char*		normalized;
 
-	PED_ASSERT (path != NULL, return);
-
+	normalized = _normalize_path (path);
+	if (!normalized)
+		return;
 	for (dev = devices; dev; dev = dev->next) {
-		if (!strcmp (dev->path, path))
+		if (!strcmp (dev->path, normalized))
 			return;
 	}
+	ped_free (normalized);
 
 	ped_exception_fetch_all ();
-
 	dev = ped_device_new (path);
 	if (!dev)
 		goto error;
-
-	if (!ped_device_open (dev)) {
-		ped_device_destroy (dev);
-		goto error;
-	}
+	if (!ped_device_open (dev))
+		goto error_destroy_dev;
 	ped_exception_leave_all ();
 
 	ped_device_close (dev);
 	ped_device_add (dev);
-
 	return;
 
+error_destroy_dev:
+	ped_device_destroy (dev);
 error:
 	ped_exception_catch ();
 	ped_exception_leave_all ();
