@@ -182,6 +182,7 @@ ped_disk_alloc (PedDevice* dev, PedDiskType* disk_type)
 	disk->dev = dev;
 	disk->type = disk_type;
 	disk->update_mode = 0;
+	disk->disk_specific = NULL;
 
 	disk->part_list = ped_partition_new (
 		disk, PED_PARTITION_FREESPACE, NULL,
@@ -251,6 +252,41 @@ ped_disk_type_check_feature (const PedDiskType* disk_type,
 	return (disk_type->features & feature) != 0;
 }
 
+int
+ped_disk_get_primary_partition_count (PedDisk* disk)
+{
+	PedPartition*	walk;
+	int		count = 0;
+
+	PED_ASSERT (disk != NULL, return 0);
+
+	for (walk = disk->part_list; walk;
+	     walk = ped_disk_next_partition (disk, walk)) {
+		if (walk->type == PED_PARTITION_PRIMARY
+		    || walk->type == PED_PARTITION_EXTENDED)
+			count++;
+	}
+
+	return count;
+}
+
+int
+ped_disk_get_last_partition_num (PedDisk* disk)
+{
+	PedPartition*	walk;
+	int		highest = -1;
+
+	PED_ASSERT (disk != NULL, return 0);
+
+	for (walk = disk->part_list; walk;
+	     walk = ped_disk_next_partition (disk, walk)) {
+		if (walk->num > highest)
+			highest = walk->num;
+	}
+
+	return highest;
+}
+
 int
 ped_disk_get_max_primary_partition_count (const PedDisk* disk)
 {
@@ -314,41 +350,6 @@ _partition_enumerate (PedPartition* part)
 	return disk_type->ops->partition_enumerate (part);
 }
 
-static int
-_disk_count_primary_partitions (PedDisk* disk)
-{
-	PedPartition*	walk;
-	int		count = 0;
-
-	PED_ASSERT (disk != NULL, return 0);
-
-	for (walk = disk->part_list; walk;
-	     walk = ped_disk_next_partition (disk, walk)) {
-		if (walk->type == PED_PARTITION_PRIMARY
-		    || walk->type == PED_PARTITION_EXTENDED)
-			count++;
-	}
-
-	return count;
-}
-
-static int
-_disk_last_partition_num (PedDisk* disk)
-{
-	PedPartition*	walk;
-	int		highest = -1;
-
-	PED_ASSERT (disk != NULL, return 0);
-
-	for (walk = disk->part_list; walk;
-	     walk = ped_disk_next_partition (disk, walk)) {
-		if (walk->num > highest)
-			highest = walk->num;
-	}
-
-	return highest;
-}
-
 /* gives all the (active) partitions a number.  It should preserve the numbers
  * and orders as much as possible.
  */
@@ -365,7 +366,7 @@ ped_disk_enumerate_partitions (PedDisk* disk)
  * is removed, then all logical partitions that were number higher MUST be
  * renumbered)
  */
-	end = _disk_last_partition_num (disk);
+	end = ped_disk_get_last_partition_num (disk);
 	for (i=1; i<=end; i++) {
 		walk = ped_disk_get_partition (disk, i);
 		if (!walk)
@@ -591,6 +592,7 @@ ped_partition_alloc (const PedDisk* disk, PedPartitionType type,
 	part->type = type;
 	part->part_list = NULL;
 	part->fs_type = fs_type;
+	part->disk_specific = NULL;
 
 	return part;
 
@@ -609,12 +611,16 @@ ped_partition_new (const PedDisk* disk, PedPartitionType type,
 	PedPartition*	part;
 
 	PED_ASSERT (disk != NULL, return NULL);
+	PED_ASSERT (disk->type != NULL, return NULL);
+	PED_ASSERT (disk->type->ops != NULL, return NULL);
 	PED_ASSERT (disk->type->ops->partition_new != NULL, return NULL);
 
 	supports_extended = ped_disk_type_check_feature (disk->type,
 			    	PED_DISK_TYPE_EXTENDED);
 
-	if (!supports_extended && type == PED_PARTITION_EXTENDED) {
+	if (!supports_extended
+	    && (type == PED_PARTITION_EXTENDED
+			|| type == PED_PARTITION_LOGICAL)) {
 		ped_exception_throw (
 			PED_EXCEPTION_ERROR,
 			PED_EXCEPTION_CANCEL,
@@ -870,24 +876,28 @@ ped_partition_get_name (const PedPartition* part)
 static int
 _partition_is_root_device (const PedPartition* part)
 {
-	char		part_name [256];
+	char*		part_name;
 	unsigned int	root_dev;
 	struct stat	part_stat;
 	FILE*		proc_real_root_dev;
 
-	sprintf (part_name, "%s%d", part->geom.disk->dev->path, part->num);
-
+	part_name = ped_device_get_part_path (part->geom.disk->dev, part->num);
+	if (!part_name)
+		return 0;
 	if (stat (part_name, &part_stat))
-		goto error;
+		goto error_free_part_name;
 
 	proc_real_root_dev = fopen ("/proc/sys/kernel/real-root-dev", "r");
 	if (!proc_real_root_dev)
-		goto error;
+		goto error_free_part_name;
 	fscanf (proc_real_root_dev, "%d", &root_dev);
 	fclose (proc_real_root_dev);
 
+	ped_free (part_name);
 	return root_dev == part_stat.st_rdev;
 
+error_free_part_name:
+	ped_free (part_name);
 error:
 	return 0;
 }
@@ -944,7 +954,7 @@ error:
 static int
 _partition_is_mounted (const PedPartition* part)
 {
-	char		part_name [256];
+	char*		part_name;
 	int		status;
 
 	PED_ASSERT (part != NULL, return 0);
@@ -952,27 +962,39 @@ _partition_is_mounted (const PedPartition* part)
 	if (part->num == -1)
 		goto error;
 
-	sprintf (part_name, "%s%d", part->geom.disk->dev->path, part->num);
+	part_name = ped_device_get_part_path (part->geom.disk->dev, part->num);
+	if (!part_name)
+		goto error;
 
 #ifdef linux
-	if (_partition_is_root_device (part))
+	if (_partition_is_root_device (part)) {
+		ped_free (part_name);
 		return 1;
+	}
 
 	status = _check_mount_table_is_busy ("/proc/mounts", part_name);
-	if (status == 1)
+	if (status == 1) {
+		ped_free (part_name);
 		return 1;
+	}
 
 	if (status != -1) {
 		status = _check_mount_table_is_busy ("/proc/swaps", part_name);
-		if (status != -1)
+		if (status != -1) {
+			ped_free (part_name);
 			return status;
+		}
 	}
 #endif
 
 	status = _check_mount_table_is_busy ("/etc/mtab", part_name);
-	if (status != -1)
+	if (status != -1) {
+		ped_free (part_name);
 		return status;
+	}
 
+error_free_part_name:
+	ped_free (part_name);
 error:
 	return ped_exception_throw (
 			PED_EXCEPTION_WARNING,
@@ -1362,10 +1384,10 @@ ped_disk_add_partition (PedDisk* disk, PedPartition* part,
 	PED_ASSERT (part != NULL, return 0);
 
 #ifdef VERBOSE
-	printf ("ped_disk_add_partition (dev=\"%s\", start=%d, end=%d,"
+	printf ("ped_disk_add_partition (dev=\"%s\", start=%llx, end=%llx,"
 		" type=%x)\n",
-		disk->dev->path, (int) part->geom.start, (int) part->geom.end,
-		(int) part->system);
+		disk->dev->path, part->geom.start, part->geom.end,
+		part->type);
 #endif
 
 	if (!ped_disk_type_check_feature (disk->type, PED_DISK_TYPE_EXTENDED)
@@ -1382,7 +1404,7 @@ ped_disk_add_partition (PedDisk* disk, PedPartition* part,
 
 	if (part->type == PED_PARTITION_PRIMARY
 	    || part->type == PED_PARTITION_EXTENDED) {
-		if (_disk_count_primary_partitions (disk) + 1
+		if (ped_disk_get_primary_partition_count (disk) + 1
 		    > ped_disk_get_max_primary_partition_count (disk)) {
 			ped_exception_throw (
 				PED_EXCEPTION_ERROR,
