@@ -111,6 +111,7 @@ static int msdos_alloc_metadata (PedDisk* disk);
 static int msdos_get_max_primary_partition_count (const PedDisk* disk);
 
 static PedSector chs_to_sector (const PedDevice* dev, const RawCHS* chs);
+static int chs_to_cylinder (const RawCHS* chs);
 
 static PedDiskOps msdos_disk_ops = {
 	probe:			msdos_probe,
@@ -174,6 +175,7 @@ check_partition_sanity (const PedDevice* dev, const DosRawPartition* raw_part)
 	PedSector	legacy_start;
 	PedSector	legacy_end;
 	PedSector	max_chs_sector;
+	PedSector	cyl_size = dev->heads * dev->sectors;
 
 	if (raw_part->type == 0) {
 		DosRawPartition		empty;
@@ -206,8 +208,17 @@ check_partition_sanity (const PedDevice* dev, const DosRawPartition* raw_part)
 	    && PED_LE32_TO_CPU (raw_part->start) > max_chs_sector)
 		return 0;
 	if (legacy_end) {
-		if (PED_LE32_TO_CPU (raw_part->length) > max_chs_sector)
-			return 0;
+		if (PED_LE32_TO_CPU (raw_part->length) > max_chs_sector) {
+			/* subtle: some broken partition program might
+			 * have naively set end cylinder to > 1024, and
+			 * got wrapped around...
+			 */
+			if (chs_to_cylinder (&raw_part->chs_end)
+					== (legacy_end / cyl_size) % 1024)
+				return 1;
+			else 
+				return 0;
+		}
 		if (PED_LE32_TO_CPU (raw_part->start)
 		    + PED_LE32_TO_CPU (raw_part->length) - 1 > max_chs_sector)
 			return 0;
@@ -221,13 +232,16 @@ msdos_probe (const PedDevice *dev)
 {
 	PedDiskType*	disk_type;
 	DosRawTable	part_table;
+	char		possible_pc98_table [PED_SECTOR_SIZE];
+	int		pc98_is_zero;
 	int		i;
 
 	PED_ASSERT (dev != NULL, return 0);
 
 	if (!ped_device_open ((PedDevice*) dev))
 		return 0;
-	if (!ped_device_read (dev, &part_table, 0, 1)) {
+	if (!ped_device_read (dev, &part_table, 0, 1)
+	    || !ped_device_read (dev, &possible_pc98_table, 1, 1)) {
 		ped_device_close ((PedDevice*) dev);
 		return 0;
 	}
@@ -237,11 +251,26 @@ msdos_probe (const PedDevice *dev)
 	if (PED_LE16_TO_CPU (part_table.magic) != MSDOS_MAGIC)
 		return 0;
 
-	/* check partition table sanity */
-	for (i = 0; i < 4; i++) {
-		if (!check_partition_sanity (dev, &part_table.partitions[i]))
-			return 0;
+#ifdef ENABLE_PC98
+	pc98_is_zero = 1;
+	for (i = 0; i < PED_SECTOR_SIZE; i++) {
+		if (possible_pc98_table[i])
+			pc98_is_zero = 0;
 	}
+	if (!pc98_is_zero) {
+		/* check partition table sanity */
+		for (i = 0; i < 4; i++) {
+			if (!check_partition_sanity (dev,
+						     &part_table.partitions[i]))
+				return 0;
+		}
+	}
+
+	/* If this is a GPT disk, fail here */
+	for (i = 0; i < 4; i++) {
+		if (part_table.partitions[i].type == 0xEE)
+ 			return 0;
+ 	}
 
 	/* HACK: it's impossible to tell PC98 and msdos disk labels apart.
 	 * Someone made the signatures the same (very clever).  Since
@@ -250,6 +279,7 @@ msdos_probe (const PedDevice *dev)
 	disk_type = ped_disk_type_get (PC98_NAME);
 	if (disk_type && disk_type->ops->probe (dev))
 		return 0;
+#endif /* ENABLE_PC98 */
 
 	return 1;
 }
@@ -286,7 +316,7 @@ error:
 static PedDisk*
 msdos_create (PedDevice* dev)
 {
-	DosRawTable		table;
+	DosRawTable	table;
 
 	PED_ASSERT (dev != NULL, return 0);
 
@@ -374,6 +404,12 @@ chs_to_sector (const PedDevice* dev, const RawCHS* chs)
 		* (dev->sector_size / 512);
 }
 
+static int
+chs_to_cylinder (const RawCHS* chs)
+{
+	return chs->cylinder + ((chs->sector >> 6) << 8);
+}
+
 static void
 sector_to_chs (const PedDevice* dev, PedSector sector, RawCHS* chs)
 {
@@ -558,7 +594,8 @@ process_inconsistent_table (PedDisk* disk, const DosRawPartition* raw_part,
 			"geometry incorrectly.  However, this does not "
 			"appear to be the case here.  It is safe to ignore,"
 			"but ignoring may cause (fixable) problems with "
-			"some boot loaders."),
+			"some boot loaders, and may cause problems with "
+			"FAT file systems.  Using LBA is recommended."),
 			disk->dev->path)
 				== PED_EXCEPTION_CANCEL)
 			return 0;
@@ -1275,6 +1312,9 @@ _logical_constraint (PedDisk* disk, PedSector start_offset, int is_start_part)
 
 		min_start = ped_round_up_to (ext_part->geom.start + 1,
 					     cylinder_size);
+		if (min_start >= ext_part->geom.end)
+			return NULL;
+
 		max_length = ext_part->geom.end - min_start + 1;
 		if (!ped_geometry_init (&max_geom, disk, min_start, max_length))
 			return NULL;
