@@ -238,19 +238,6 @@ sun_probe (const PedDevice *dev)
 		return 0;
 	}
 
-	/* We have to fail here. Presumably, this is either a badly
-	   created label, or the detected geometry isn't correct. Either
-	   way, we do not want to even attempt to mess with this. The user
-	   will have to resolve this themselves somehow.  */
-	if (PED_BE16_TO_CPU(label.nsect) != dev->sectors ||
-	    PED_BE16_TO_CPU(label.ntrks) != dev->heads) {
-		ped_exception_throw (
-				PED_EXCEPTION_ERROR,
-				PED_EXCEPTION_CANCEL,
-				_("Disk geometry does not match disk label."));
-		return 0;
-	}
-
 	return 1;
 }
 
@@ -310,7 +297,7 @@ sun_create (PedDevice* dev)
 	label.infos[WHOLE_DISK_PART].id = WHOLE_DISK_ID;
 	label.partitions[WHOLE_DISK_PART].start_cylinder = 0;
 	label.partitions[WHOLE_DISK_PART].num_sectors =
-		PED_CPU_TO_BE32(label.ntrks * label.nsect * label.ncyl);
+		PED_CPU_TO_BE32((dev->cylinders-2) * dev->heads * dev->sectors);
 
 	/* Now a neato string to describe this label */
 	snprintf(label.info, sizeof(label.info) - 1,
@@ -360,6 +347,43 @@ error:
 	return 0;
 }
 
+static int
+_check_geometry_sanity (PedDisk* disk, SunRawLabel* label)
+{
+	PedDevice*	dev = disk->dev;
+
+	if (PED_BE16_TO_CPU(label->nsect) != dev->sectors ||
+	    PED_BE16_TO_CPU(label->ntrks) != dev->heads) {
+		if (ped_exception_throw (
+				PED_EXCEPTION_WARNING,
+				PED_EXCEPTION_IGNORE_CANCEL,
+				_("The disk CHS geometry (%d,%d,%d) does not "
+				  "match the geometry stored on the disk "
+				  "label (%d,%d,%d)."),
+				dev->cylinders, dev->heads, dev->sectors,
+				PED_BE16_TO_CPU(label->pcylcount),
+				PED_BE16_TO_CPU(label->ntrks),
+				PED_BE16_TO_CPU(label->nsect))
+			== PED_EXCEPTION_CANCEL)
+			return 0;
+		dev->sectors = PED_BE16_TO_CPU(label->nsect);
+		dev->heads = PED_BE16_TO_CPU(label->ntrks);
+		dev->cylinders = PED_BE16_TO_CPU(label->pcylcount);
+
+		if (dev->sectors * dev->heads * dev->cylinders > dev->length) {
+			if (ped_exception_throw (
+				PED_EXCEPTION_WARNING,
+				PED_EXCEPTION_IGNORE_CANCEL,
+				_("The disk label describes a disk bigger than "
+				  "%s."),
+				dev->path)
+				!= PED_EXCEPTION_IGNORE)
+				return 0;
+		}
+	}
+	return 1;
+}
+
 static int
 sun_read (PedDisk* disk)
 {
@@ -382,6 +406,9 @@ sun_read (PedDisk* disk)
 	if (!ped_device_read (disk->dev, &label, 0, 1))
 		goto error_close_dev;
 
+	if (!_check_geometry_sanity (disk, &label))
+		goto error_close_dev;
+
 	disk_data = disk->disk_specific;
 	block = disk->dev->sectors * disk->dev->heads;
 	disk_data->length = block * (disk->dev->cylinders
@@ -638,7 +665,7 @@ sun_get_max_primary_partition_count (const PedDisk* disk)
 }
 
 static PedConstraint*
-_primary_constraint (PedDisk* disk)
+_get_strict_constraint (PedDisk* disk)
 {
         PedAlignment    start_align;
         PedAlignment    end_align;
@@ -657,6 +684,23 @@ _primary_constraint (PedDisk* disk)
                                    &max_geom, 1);
 }
 
+static PedConstraint*
+_get_lax_constraint (PedDisk* disk)
+{
+        PedAlignment    start_align;
+        PedGeometry     max_geom;
+	SunDiskData*	disk_data = disk->disk_specific;
+	PedSector	block = disk->dev->sectors * disk->dev->heads;
+
+        if (!ped_alignment_init (&start_align, 0, block))
+                return NULL;
+	if (!ped_geometry_init (&max_geom, disk, 0, disk_data->length))
+		return NULL;
+
+        return ped_constraint_new (&start_align, ped_alignment_any, &max_geom,
+                                   &max_geom, 1);
+}
+
 /* argghh!  this should be somewhere else... but where?! */
 static int
 _try_constraint (PedPartition* part, const PedConstraint* external,
@@ -684,21 +728,31 @@ fail:
         return 0;
 }
 
+/* _get_strict_constraint() will align the partition to the end of the cylinder.
+ * This isn't required, but since partitions must start at the start of the
+ * cylinder, space between the end of a partition and the end of a cylinder
+ * is unusable, so there's no point wasting space!
+ *	However, if they really insist (via constraint)... which they will
+ * if they're reading a weird table of the disk... then we allow the end to
+ * be anywhere, with _get_lax_constraint()
+ */
 static int
 sun_partition_align (PedPartition* part, const PedConstraint* constraint)
 {
         PED_ASSERT (part != NULL, return 0);
 
-        if (!_try_constraint (part, constraint,
-                             _primary_constraint (part->geom.disk))) {
-                ped_exception_throw (
-                        PED_EXCEPTION_ERROR,
-                        PED_EXCEPTION_IGNORE_CANCEL,
-                        _("Unable to align partition."));
-                return 0;
-        }
+        if (_try_constraint (part, constraint,
+                             _get_strict_constraint (part->geom.disk)))
+		return 1;
+        if (_try_constraint (part, constraint,
+                             _get_lax_constraint (part->geom.disk)))
+        	return 1;
 
-        return 1;
+	ped_exception_throw (
+		PED_EXCEPTION_ERROR,
+		PED_EXCEPTION_CANCEL,
+		_("Unable to align partition."));
+	return 0;
 }
 
 static int
