@@ -25,6 +25,7 @@ static const char _ext2_c[] = "$Id: ext2.c,v 1.54 1999/10/08 06:20:16 buytenh Ex
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
+#include <uuid/uuid.h>
 #include "ext2.h"
 
 /* ext2 stuff ****************************************************************/
@@ -367,13 +368,15 @@ int ext2_set_inode_state(struct ext2_fs *fs, ino_t inode, int state, int updatem
 static void
 _inode_update_size(struct ext2_fs *fs, struct ext2_inode *inode, int delta)
 {
-	int		 i512perblock = 1 << (fs->logsize - 9);
+	int		i512perblock = 1 << (fs->logsize - 9);
+	__u64		size;
 
 	/* i_blocks is in 512 byte blocks */
 	inode->i_blocks = PED_CPU_TO_LE32(EXT2_INODE_BLOCKS(*inode)
 		       			  + delta * i512perblock);
-	inode->i_size = PED_CPU_TO_LE32(EXT2_INODE_SIZE(*inode)
-		       			+ delta * fs->blocksize);
+	size = EXT2_INODE_SIZE(*inode) + delta * fs->blocksize;
+	inode->i_size = PED_CPU_TO_LE32(size % 0x100000000);
+	inode->i_size_high = PED_CPU_TO_LE32(size / 0x100000000);
 	inode->i_mtime = PED_CPU_TO_LE32(time(NULL));
 }
 
@@ -604,10 +607,6 @@ int ext2_is_group_sparse(struct ext2_fs *fs, int group)
 
 void ext2_close(struct ext2_fs *fs)
 {
-#if 0
-	ext2_journal_deinit(fs);
-#endif
-
 	ext2_commit_metadata(fs, EXT2_META_PRIMARY | EXT2_META_BACKUP);
 	ext2_sync(fs);
 
@@ -757,24 +756,33 @@ struct ext2_fs *ext2_open(struct ext2_dev_handle *handle, int state)
 
 	if (EXT2_SUPER_STATE(fs->sb) & EXT2_ERROR_FS & ~(state & EXT2_ERROR_FS))
 	{
-		ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
-		     _("Filesystem has errors!  Run e2fsck first!"));
-		goto error_free_fs;
+		if (ped_exception_throw (
+			PED_EXCEPTION_WARNING, PED_EXCEPTION_IGNORE_CANCEL,
+			_("Filesystem has errors!  You should run e2fsck."))
+				== PED_EXCEPTION_CANCEL)
+			goto error_free_fs;
 	}
 
-	if (!((EXT2_SUPER_STATE(fs->sb) | state) & EXT2_VALID_FS))
+	if (!((EXT2_SUPER_STATE(fs->sb) | state) & EXT2_VALID_FS)
+	    || (EXT2_SUPER_FEATURE_INCOMPAT(fs->sb)
+	    	& EXT3_FEATURE_INCOMPAT_RECOVER))
 	{
-		ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
-		     _("Filesystem was not cleanly unmounted!  "
-		       "Run e2fsck first!"));
-		goto error_free_fs;
+		if (ped_exception_throw (
+			PED_EXCEPTION_WARNING, PED_EXCEPTION_IGNORE_CANCEL,
+			_("Filesystem was not cleanly unmounted!  "
+			  "You should e2fsck."))
+				== PED_EXCEPTION_CANCEL)
+			goto error_free_fs;
 	}
 
-	if (EXT2_SUPER_FEATURE_COMPAT(fs->sb) ||
+	if ((EXT2_SUPER_FEATURE_COMPAT(fs->sb)
+	                & ~EXT3_FEATURE_COMPAT_HAS_JOURNAL) ||
 	    (EXT2_SUPER_FEATURE_INCOMPAT(fs->sb)
-	    		& ~EXT2_FEATURE_INCOMPAT_FILETYPE) ||
+	    		& ~(EXT2_FEATURE_INCOMPAT_FILETYPE |
+			    EXT3_FEATURE_INCOMPAT_RECOVER)) ||
 	    (EXT2_SUPER_FEATURE_RO_COMPAT(fs->sb)
-	     		& ~EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER))
+	                & ~(EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER |
+	    		    EXT2_FEATURE_RO_COMPAT_LARGE_FILE)))
 	{
 		ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
 		     _("Filesystem has incompatible feature enabled"));
@@ -811,6 +819,13 @@ struct ext2_fs *ext2_open(struct ext2_dev_handle *handle, int state)
 			& EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER)
 		fs->sparse = 1;
 
+	fs->has_journal = 0 < (EXT2_SUPER_FEATURE_COMPAT(fs->sb)
+	    			& EXT3_FEATURE_COMPAT_HAS_JOURNAL);
+	fs->has_internal_journal
+		= fs->has_journal
+			&& uuid_is_null(EXT2_SUPER_JOURNAL_UUID(fs->sb))
+			&& EXT2_SUPER_JOURNAL_INUM(fs->sb);
+
 	fs->gd = ped_malloc (fs->numgroups * sizeof (struct ext2_group_desc)
 			     	+ fs->blocksize);
 	if (!fs->gd)
@@ -824,10 +839,6 @@ struct ext2_fs *ext2_open(struct ext2_dev_handle *handle, int state)
 
 	fs->metadirty = 0;
 
-#if 0
-	ext2_journal_init(fs);
-#endif
-
 	return fs;
 
 error_free_gd:
