@@ -39,6 +39,10 @@ static const char _ext2_h[] = "$Id: ext2.h,v 1.6 1999/10/02 20:56:47 aclausen Ex
 #define howmany(x,y)        (((x)+((y)-1))/(y))
 #endif
 
+/* Ehrm.... sorry, pedanticists! :-) */
+#define offsetof(type, field) ((size_t)(&(((type *)0)->field)))
+
+
 #ifndef EXT2_RESIZE_INO
 #define EXT2_RESIZE_INO 7
 #endif
@@ -109,6 +113,8 @@ struct ext2_fs
 									   2:only first sb&gd copy clean */
 
 	int				  sparse;			/* sparse superblocks */
+	int				  has_journal;			/* journal */
+	int				  has_internal_journal;
 
 	int				  blocksize;
 	int				  logsize;
@@ -125,8 +131,6 @@ struct ext2_fs
 	int				 opt_debug;
 	int				 opt_safe;
 	int				 opt_verbose;
-
-	void				 *journal;
 };
 
 
