@@ -570,19 +570,30 @@ int ext2_resize_fs(struct ext2_fs *fs, blk_t newsize)
 	blk_t residue;
 	int status;
 
+	if (EXT2_SUPER_STATE(fs->sb) & EXT2_ERROR_FS)
+	{
+		ped_exception_throw (
+			PED_EXCEPTION_WARNING, PED_EXCEPTION_CANCEL,
+			_("Filesystem has errors!  You should run e2fsck."));
+		return 0;
+	}
+
+	if (!(EXT2_SUPER_STATE(fs->sb) & EXT2_VALID_FS))
+	{
+		ped_exception_throw (
+			PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
+			_("Filesystem was not cleanly unmounted!  "
+			  "You should e2fsck."));
+		return 0;
+	}
+
 	if (fs->opt_verbose)
 		fprintf(stderr, "ext2_resize_fs\n");
 
 	residue = (newsize - EXT2_SUPER_FIRST_DATA_BLOCK(fs->sb))
 		   % EXT2_SUPER_BLOCKS_PER_GROUP(fs->sb);
 	if (residue && residue <= fs->adminblocks)
-	{
-/*
-		fprintf(stderr, "ext2_resize_fs: %i is an impossible size for an ext2 fs! rounding down to %i\n",
-			newsize, newsize-residue);
-*/
 		newsize -= residue;
-	}
 
 	if (newsize == EXT2_SUPER_BLOCKS_COUNT(fs->sb))
 		return 1;
