@@ -31,6 +31,7 @@ static const char _interface_c[] = "$Id: interface.c,v 1.6 1999/10/02 20:56:47 a
 #include <parted/disk_pc98.h>
 #include <parted/disk_mac.h>
 #include <parted/disk_bsd.h>
+#include <parted/disk_gpt.h>
 #include <parted/disk_sun.h>
 
 #include <string.h>
@@ -45,6 +46,9 @@ struct ext2_dev_handle* ext2_make_dev_handle_from_parted_geometry(PedGeometry* g
 static int
 _ext2_probe (const PedGeometry* geom);
 
+static int
+_ext3_probe (const PedGeometry* geom);
+
 static int
 _ext2_clobber (PedGeometry* geom);
 
@@ -92,30 +96,67 @@ static PedFileSystemType _ext2_type = {
        name:		 "ext2"
 };
 
+
+
+static PedFileSystemOps _ext3_ops = {
+	probe:		_ext3_probe,
+	clobber:	_ext2_clobber,
+	open:		_ext2_open,
+	create:         NULL,			/* FIXME */
+	close:		_ext2_close,
+	check:          _ext2_check,
+	resize:		_ext2_resize,
+	copy:           NULL,
+	get_resize_constraint:	_ext2_get_resize_constraint,
+	set_system:	_ext2_set_system
+};
+
+static PedFileSystemType _ext3_type = {
+       next:		 NULL,
+       ops:		 &_ext3_ops,
+       name:		 "ext3"
+};
+
 void ped_file_system_ext2_init ()
 {
+	ped_file_system_type_register (&_ext3_type);
 	ped_file_system_type_register (&_ext2_type);
 }
 
 void ped_file_system_ext2_done ()
 {
+	ped_file_system_type_unregister (&_ext3_type);
 	ped_file_system_type_unregister (&_ext2_type);
 }
 
-
-
 static int
 _ext2_probe (const PedGeometry* geom)
 {
 	struct ext2_super_block sb;
 
 	if (!ped_geometry_read(geom, &sb, 2, 2)
-	    || EXT2_SUPER_MAGIC(sb) != EXT2_SUPER_MAGIC_CONST)
+	     || EXT2_SUPER_MAGIC(sb) != EXT2_SUPER_MAGIC_CONST
+	     || EXT2_SUPER_FEATURE_COMPAT(sb) & EXT3_FEATURE_COMPAT_HAS_JOURNAL)
 		return 0;
 
 	return 1;
 }
 
+
+static int
+_ext3_probe (const PedGeometry* geom)
+{
+	struct ext2_super_block sb;
+
+	if (!ped_geometry_read(geom, &sb, 2, 2)
+	    || EXT2_SUPER_MAGIC(sb) != EXT2_SUPER_MAGIC_CONST)
+		return 0;
+	if (EXT2_SUPER_FEATURE_COMPAT(sb) & EXT3_FEATURE_COMPAT_HAS_JOURNAL)
+	        return 1;
+
+	return 0;
+}
+
 static int
 _ext2_clobber (PedGeometry* geom)
 {
@@ -236,10 +277,11 @@ _ext2_resize (PedFileSystem* fs, const PedGeometry* geom)
 		return 0;
 	}
 
-	geom->disk->dev->boot_dirty = 1;
-
+	
 	f = (struct ext2_fs *) fs->type_specific;
 
+	geom->disk->dev->boot_dirty = 1;
+	
 /* ensure that the geometry contains the new and old geometry */
 	if (old_length > geom->length) {
 		if (!ext2_resize_fs(f, geom->length >> (f->logsize - 9)))
@@ -349,6 +391,15 @@ _ext2_set_system (const PedFileSystem* fs, PedPartition* part,
 		return 1;
 	}
 
+	if (strcmp (disk_type->name, GPT_NAME) == 0) {
+		GPTPartitionData *gpt_part_data = part->disk_specific;
+		PED_ASSERT(gpt_part_data      != NULL, return 0);	
+		PED_ASSERT(gpt_part_data->pte != NULL, return 0);
+		gpt_part_data->pte->PartitionTypeGuid =
+			PARTITION_BASIC_DATA_GUID;
+		return 1;
+	}
+
 	if (strcmp (disk_type->name, SUN_NAME) == 0) {
 		SunPartitionData* sun_data = part->disk_specific;
 
