@@ -1,6 +1,6 @@
 /*
     libparted
-    Copyright (C) 1998-2000  Andrew Clausen  <clausen@gnu.org>
+    Copyright (C) 1998, 1999, 2000 Free Software Foundation, Inc.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -104,7 +104,7 @@ print_chain (PedFileSystem* fs, FatCluster start)
 	this_row = 0;
 	for (clst = start; !fat_table_is_eof (fs_info->fat, clst);
 	     clst = fat_table_get (fs_info->fat, clst)) {
-		printf ("  %d", clst);
+		printf ("  %d", (int) clst);
 		if (++this_row == 7) {
 			printf ("\n");
 			this_row = 0;
@@ -136,6 +136,7 @@ flag_traverse_fat (PedFileSystem* fs, const char* chain_name, FatCluster start,
 	FatCluster	clst;
 	FatCluster	prev_clst;
 	int		last_cluster_usage;
+	FatCluster	chain_length = 0;
 
 	if (fat_table_is_eof (fs_info->fat, start)) {
 		if (ped_exception_throw (
@@ -150,10 +151,12 @@ flag_traverse_fat (PedFileSystem* fs, const char* chain_name, FatCluster start,
 
 	for (prev_clst = clst = start; !fat_table_is_eof (fs_info->fat, clst);
 	     prev_clst = clst, clst = fat_table_get (fs_info->fat, clst)) {
+		chain_length++;
 		if (!clst) {
 			ped_exception_throw (PED_EXCEPTION_FATAL,
 				PED_EXCEPTION_CANCEL,
-				_("Bad FAT: unterminated chain for %s"),
+				_("Bad FAT: unterminated chain for %s.  You "
+				  "should run dosfsck or scandisk."),
 				chain_name);
 			return 0;
 		}
@@ -162,7 +165,8 @@ flag_traverse_fat (PedFileSystem* fs, const char* chain_name, FatCluster start,
 			ped_exception_throw (PED_EXCEPTION_FATAL,
 				PED_EXCEPTION_CANCEL,
 				_("Bad FAT: cluster %d outside filesystem "
-				  "in chain for %s"),
+				  "in chain for %s.  You should run dosfsck "
+				  "or scandisk."),
 				(int) clst, chain_name);
 			return 0;
 		}
@@ -170,7 +174,8 @@ flag_traverse_fat (PedFileSystem* fs, const char* chain_name, FatCluster start,
 		if (fs_info->cluster_info [clst].flag != FAT_FLAG_FREE ) {
 			ped_exception_throw (PED_EXCEPTION_FATAL,
 				PED_EXCEPTION_CANCEL,
-				_("Bad FAT: cluster %d is cross-linked for %s"),
+				_("Bad FAT: cluster %d is cross-linked for "
+				  "%s.  You should run dosfsck or scandisk."),
 				(int) clst, chain_name);
 			return 0;
 		}
@@ -182,6 +187,20 @@ flag_traverse_fat (PedFileSystem* fs, const char* chain_name, FatCluster start,
 		fs_info->cluster_info [clst].units_used = 64;
 	}
 
+	if (size
+	    && chain_length != div_round_up (size, fs_info->cluster_sectors)) {
+		if (ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_IGNORE_CANCEL,
+			_("%s is %dk, but it has is %d clusters (%dk)."),
+			chain_name,
+			(int) size / 2,
+			(int) chain_length,
+			(int) chain_length * fs_info->cluster_sectors / 2)
+				!= PED_EXCEPTION_IGNORE)
+			return 0;
+	}
+
 	last_cluster_usage
 		= div_round_up (64 * remainder_round_up (size,
 						fs_info->cluster_sectors),
@@ -199,7 +218,6 @@ flag_traverse_fat (PedFileSystem* fs, const char* chain_name, FatCluster start,
 static int
 flag_traverse_dir (FatTraverseInfo* trav_info) {
 	PedFileSystem*		fs = trav_info->fs;
-	FatSpecific*		fs_info = FAT_SPECIFIC (fs);
 	FatDirEntry*		this_entry;
 	FatTraverseInfo*	subdir_trav_info;
 	char			file_name [512];
@@ -208,10 +226,14 @@ flag_traverse_dir (FatTraverseInfo* trav_info) {
 	PedSector		size;
 	PedExceptionOption	ex_status;
 
+	PED_ASSERT (trav_info != NULL, return 0);
+
 	strcpy (file_name, trav_info->dir_name);
 	file_name_start = file_name + strlen (file_name);
 
 	while ( (this_entry = fat_traverse_next_dir_entry (trav_info)) ) {
+		if (fat_dir_entry_is_null_term (this_entry))
+			break;
 		if (!fat_dir_entry_has_first_cluster (this_entry, fs))
 			continue;
 		if (this_entry->name [0] == '.')
@@ -227,6 +249,7 @@ flag_traverse_dir (FatTraverseInfo* trav_info) {
 		print_chain (fs, first_cluster);
 #endif
 
+/*
 		if (fat_dir_entry_is_system_file (this_entry)
 		    && !is_movable_system_file (file_name)) {
 			ex_status = ped_exception_throw (
@@ -246,6 +269,7 @@ flag_traverse_dir (FatTraverseInfo* trav_info) {
 				case PED_EXCEPTION_IGNORE:
 			}
 		}
+*/
 
 		if (fat_dir_entry_is_directory (this_entry)) {
 			if (!flag_traverse_fat (fs, file_name, first_cluster,
@@ -254,6 +278,8 @@ flag_traverse_dir (FatTraverseInfo* trav_info) {
 
 			subdir_trav_info = fat_traverse_directory (trav_info,
 								   this_entry);
+			if (!subdir_trav_info)
+				return 0;
 			if (!flag_traverse_dir (subdir_trav_info))
 				return 0;
 		} else if (fat_dir_entry_is_file (this_entry)) {
@@ -267,13 +293,24 @@ flag_traverse_dir (FatTraverseInfo* trav_info) {
 	return 1;
 }
 
+static void
+_mark_bad_clusters (PedFileSystem* fs)
+{
+	FatSpecific*	fs_info = FAT_SPECIFIC (fs);
+	FatCluster	cluster;
+	FatFragment	frag;
+
+	for (cluster = 2; cluster < fs_info->cluster_count + 2; cluster++) {
+		if (fat_table_is_bad (fs_info->fat, cluster))
+			fs_info->cluster_info [cluster].flag = FAT_FLAG_BAD;
+	}
+}
+
 /*  
     fills in cluster_info.  Each FAT entry (= cluster) is flagged as either
     FAT_FLAG_FREE, FAT_FLAG_FILE or FAT_FLAG_DIRECTORY.
 
     Also, the fraction of each cluster (x/64) is recorded
-
-FIXME: mark FAT_FLAG_BAD
 */
 int
 fat_collect_cluster_info (PedFileSystem* fs) {
@@ -290,14 +327,15 @@ fat_collect_cluster_info (PedFileSystem* fs) {
 		if (!flag_traverse_dir (trav_info))
 			return 0;
 		if (!flag_traverse_fat (fs, "\\", fs_info->root_cluster,
-                                        FAT_FLAG_DIRECTORY,
-					fs_info->cluster_size))
+                                        FAT_FLAG_DIRECTORY, 0))
 			return 0;
 	} else {
 		trav_info = fat_traverse_begin (fs, FAT_ROOT, "\\");
 		if (!flag_traverse_dir (trav_info))
 			return 0;
 	}
+
+	_mark_bad_clusters (fs);
 	return 1;
 }
 
@@ -338,14 +376,14 @@ fat_get_fragment_flag (PedFileSystem* fs, FatFragment frag)
 		    return 0);
 
 	flag = fat_get_cluster_flag (fs, cluster);
-	if (flag != FAT_FLAG_FILE)
+	if (flag != FAT_FLAG_FILE && flag != FAT_FLAG_DIRECTORY)
 		return flag;
-	last_frag_used = fat_get_cluster_usage (fs, cluster)
-				/ fs_info->cluster_frags;
+	last_frag_used = (fat_get_cluster_usage (fs, cluster) - 1)
+				/ fs_info->frag_sectors;
 	if (offset > last_frag_used)
 		return FAT_FLAG_FREE;
 	else
-		return FAT_FLAG_FILE;
+		return flag;
 }
 
 int
