@@ -1,6 +1,6 @@
 /*
     libparted
-    Copyright (C) 1998-2000  Andrew Clausen  <clausen@gnu.org>
+    Copyright (C) 1998, 1999, 2000 Free Software Foundation, Inc.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -33,10 +33,9 @@
 #include <stdarg.h>
 #include <string.h>
 
-/*
-    Recursively duplicates a directory tree, modified for the new file
-    locations.
-*/
+/* Recursively builds (i.e. makes consistent) the duplicated directory tree
+ * (leaving the original directory tree in tact)
+ */
 static int
 fat_construct_directory (FatOpContext* ctx, FatTraverseInfo* trav_info)
 {
@@ -45,6 +44,8 @@ fat_construct_directory (FatOpContext* ctx, FatTraverseInfo* trav_info)
 	FatCluster		old_first_cluster;
 
 	while ( (dir_entry = fat_traverse_next_dir_entry (trav_info)) ) {
+		if (fat_dir_entry_is_null_term (dir_entry))
+			break;
 		if (!fat_dir_entry_has_first_cluster (dir_entry, ctx->old_fs))
 			continue;
 
@@ -65,6 +66,11 @@ fat_construct_directory (FatOpContext* ctx, FatTraverseInfo* trav_info)
 				return 0;
 		}
 	}
+	/* remove "stale" entries at the end */
+	while ((dir_entry = fat_traverse_next_dir_entry (trav_info))) {
+		memset (dir_entry, 0, sizeof (FatDirEntry));
+		fat_traverse_mark_dirty (trav_info);
+	}
 	fat_traverse_complete (trav_info);
 	return 1;
 }
@@ -138,6 +144,11 @@ fat_convert_directory (FatOpContext* ctx, FatTraverseInfo* old_trav,
 	FatCluster		old_first_cluster;
 
 	while ( (old_dir_entry = fat_traverse_next_dir_entry (old_trav)) ) {
+		if (fat_dir_entry_is_null_term (old_dir_entry))
+			break;
+		if (!fat_dir_entry_is_active (old_dir_entry))
+			continue;
+
 		new_dir_entry = fat_traverse_next_dir_entry (new_trav);
 		if (!new_dir_entry) {
 			return ped_exception_throw (PED_EXCEPTION_ERROR,
@@ -148,9 +159,6 @@ fat_convert_directory (FatOpContext* ctx, FatTraverseInfo* old_trav,
 					== PED_EXCEPTION_IGNORE;
 		}
 
-		if (!fat_dir_entry_is_active (old_dir_entry))
-			continue;
-
 		*new_dir_entry = *old_dir_entry;
 		fat_traverse_mark_dirty (new_trav);
 
@@ -178,6 +186,14 @@ fat_convert_directory (FatOpContext* ctx, FatTraverseInfo* old_trav,
 		}
 	}
 
+	/* remove "stale" entries at the end, just in case there is some
+	 * overlap
+	 */
+	while ((new_dir_entry = fat_traverse_next_dir_entry (new_trav))) {
+		memset (new_dir_entry, 0, sizeof (FatDirEntry));
+		fat_traverse_mark_dirty (new_trav);
+	}
+
 	fat_traverse_complete (old_trav);
 	fat_traverse_complete (new_trav);
 	return 1;
@@ -200,13 +216,19 @@ clear_cluster (PedFileSystem* fs, FatCluster cluster)
 static int
 alloc_root_dir (FatOpContext* ctx)
 {
+	FatSpecific*		old_fs_info = FAT_SPECIFIC (ctx->old_fs);
 	FatSpecific*		new_fs_info = FAT_SPECIFIC (ctx->new_fs);
-	int			i;
+	FatCluster		i;
 	FatCluster		cluster;
+	FatCluster		cluster_count;
 
 	PED_ASSERT (new_fs_info->fat_type == FAT_TYPE_FAT32, return 0);
 
-	for (i=0; i<16; i++) {
+	cluster_count = div_round_up (
+			PED_MAX (16, old_fs_info->root_dir_sector_count),
+			new_fs_info->cluster_sectors);
+
+	for (i = 0; i < cluster_count; i++) {
 		cluster = fat_table_alloc_check_cluster (new_fs_info->fat,
 							 ctx->new_fs);
 		if (!cluster)
@@ -411,12 +433,115 @@ fat_construct_new_fat (FatOpContext* ctx)
 	return 1;
 }
 
+static int
+ask_type (PedFileSystem* fs, int fat16_ok, int fat32_ok, FatType* out_fat_type)
+{
+	FatSpecific*		fs_info = FAT_SPECIFIC (fs);
+	PedExceptionOption	status;
+	char*			fat16_msg;
+	char*			fat32_msg;
+
+	if (fs_info->fat_type == FAT_TYPE_FAT16)
+		fat16_msg = _("If leave your file system as FAT16, "
+			      "then you will have no problems.");
+	else
+		fat16_msg = _("If you convert to FAT16, and MS Windows "
+			      "is installed on this partition, then "
+			      "you must re-install the MS Windows boot "
+			      "loader.  If you want to do this, you "
+			      "should consult the Parted manual (or "
+			      "your distribution's manual).");
+
+	if (fs_info->fat_type == FAT_TYPE_FAT32)
+		fat32_msg = _("If you leave your file system as FAT32, "
+			      "then you will not introduce any new "
+			      "problems.");
+	else
+		fat32_msg = _("If you convert to FAT32, and MS Windows "
+			      "is installed on this partition, then "
+			      "you must re-install the MS Windows boot "
+			      "loader.  If you want to do this, you "
+			      "should consult the Parted manual (or "
+			      "your distribution's manual).  Also, "
+			      "converting to FAT32 will make the file "
+			      "system unreadable by MS DOS, MS Windows "
+			      "95a, and MS Windows NT.");
+
+	if (fat16_ok && fat32_ok) {
+		status = ped_exception_throw (
+			 PED_EXCEPTION_INFORMATION,
+			 PED_EXCEPTION_YES_NO_CANCEL,
+			 _("%s  %s  %s"),
+			 _("Would you like to use FAT32?"),
+			 fat16_msg,
+			 fat32_msg);
+
+		switch (status) {
+		case PED_EXCEPTION_YES:
+			*out_fat_type = FAT_TYPE_FAT32;
+			return 1;
+
+		case PED_EXCEPTION_NO:
+			*out_fat_type = FAT_TYPE_FAT16;
+			return 1;
+
+		case PED_EXCEPTION_UNHANDLED:
+			*out_fat_type = fs_info->fat_type;
+			return 1;
+
+		case PED_EXCEPTION_CANCEL:
+			return 0;
+		}
+	}
+
+	if (fat16_ok) {
+		if (fs_info->fat_type != FAT_TYPE_FAT16) {
+			status = ped_exception_throw (
+				PED_EXCEPTION_WARNING,
+				PED_EXCEPTION_OK_CANCEL,
+				_("%s  %s"),
+				_("The file system can only be resized to this "
+				  "size by converting to FAT16."),
+				fat16_msg);
+			if (status == PED_EXCEPTION_CANCEL)
+				return 0;
+		}
+		*out_fat_type = FAT_TYPE_FAT16;
+		return 1;
+	}
+
+	if (fat32_ok) {
+		if (fs_info->fat_type != FAT_TYPE_FAT32) {
+			status = ped_exception_throw (
+				PED_EXCEPTION_WARNING,
+				PED_EXCEPTION_OK_CANCEL,
+				_("%s  %s"),
+				_("The file system can only be resized to this "
+				  "size by converting to FAT32."),
+				fat32_msg);
+			if (status == PED_EXCEPTION_CANCEL)
+				return 0;
+		}
+		*out_fat_type = FAT_TYPE_FAT32;
+		return 1;
+	}
+	
+	ped_exception_throw (
+		PED_EXCEPTION_NO_FEATURE,
+		PED_EXCEPTION_CANCEL,
+		_("GNU Parted can not resize this partition to this size.  "
+		"We're working on it!"));
+
+	return 0;
+}
+
 /*  For resize operations: determine if the file system must be FAT16 or FAT32,
  *  or either.  If the new file system must be FAT32, then query for
  *  confirmation.  If either filesystem can be used, query for which one.
  */
 static int
-query_fat_type (PedFileSystem* fs, PedGeometry* new_geom, FatType* out_fat_type)
+get_fat_type (PedFileSystem* fs, const PedGeometry* new_geom,
+	      FatType* out_fat_type)
 {
 	FatSpecific*		fs_info = FAT_SPECIFIC (fs);
 	PedExceptionOption	status;
@@ -447,63 +572,7 @@ query_fat_type (PedFileSystem* fs, PedGeometry* new_geom, FatType* out_fat_type)
 				    &dummy_cluster_count,
 				    &dummy_fat_sectors);
 
-	if (fat16_ok && fat32_ok) {
-		status = ped_exception_throw (
-			 PED_EXCEPTION_INFORMATION,
-			 PED_EXCEPTION_YES_NO_CANCEL,
-			 _("Would you like to use FAT32?  It is more "
-			 "efficient, but it is not compatible with MS-DOS, "
-			 "early versions of MS-Windows 95 and Windows NT.  If "
-			 "you use these operating systems, then select No.  If "
-			 "you only use Linux, BSD, MS Windows 98 and/or MS "
-			 "Windows 95 B, then you may select Yes."));
-
-		switch (status) {
-		case PED_EXCEPTION_YES:
-			*out_fat_type = FAT_TYPE_FAT32;
-			return 1;
-
-		case PED_EXCEPTION_NO:
-			*out_fat_type = FAT_TYPE_FAT16;
-			return 1;
-
-		case PED_EXCEPTION_CANCEL:
-			return 0;
-		}
-	}
-
-	if (fat16_ok) {
-		*out_fat_type = FAT_TYPE_FAT16;
-		return 1;
-	}
-
-	if (fat32_ok) {
-		*out_fat_type = FAT_TYPE_FAT32;
-		if (fs_info->fat_type == FAT_TYPE_FAT32)
-			return 1;
-
-		status = ped_exception_throw (
-				PED_EXCEPTION_INFORMATION,
-				PED_EXCEPTION_OK_CANCEL,
-				_("The filesystem is going to be too big "
-				"for FAT16, so FAT32 must be used.  This is "
-				"not compatible with MS-DOS, early versions of "
-				"MS-Windows 95 and Windows NT.  If you use "
-				"these operating systems, then select Cancel, "
-				"and create a smaller partition.  If you only "
-				"use Linux, BSD, MS Windows 98 and/or MS "
-				"Windows 95 B, then select OK."));
-
-		return status == PED_EXCEPTION_OK;
-	}
-
-	ped_exception_throw (
-		PED_EXCEPTION_NO_FEATURE,
-		PED_EXCEPTION_CANCEL,
-		_("GNU Parted can not resize this partition to this size.  "
-		"We're working on it!"));
-
-	return 0;
+	return ask_type (fs, fat16_ok, fat32_ok, out_fat_type);
 }
 
 /*  Creates the PedFileSystem struct for the new resized file system, and
@@ -511,7 +580,7 @@ query_fat_type (PedFileSystem* fs, PedGeometry* new_geom, FatType* out_fat_type)
     (ctx->old_fs) is destroyed, and replaced with the new one (ctx->new_fs).
  */
 static FatOpContext*
-create_resize_context (PedFileSystem* fs, PedGeometry* new_geom)
+create_resize_context (PedFileSystem* fs, const PedGeometry* new_geom)
 {
 	FatSpecific*	fs_info = FAT_SPECIFIC (fs);
 	FatSpecific*	new_fs_info;
@@ -532,7 +601,7 @@ create_resize_context (PedFileSystem* fs, PedGeometry* new_geom)
 		root_dir_sector_count = FAT_ROOT_DIR_ENTRY_COUNT
 						* sizeof (FatDirEntry) / 512;
 
-	if (!query_fat_type (fs, new_geom, &new_fat_type))
+	if (!get_fat_type (fs, new_geom, &new_fat_type))
 		return 0;
 
 	fat_calc_resize_sizes (new_geom, fs_info->cluster_sectors, new_fat_type,
@@ -666,8 +735,39 @@ resize_context_abort (FatOpContext* ctx)
 	return 1;
 }
 
+/* copies the "hidden" sectors, between the boot sector and the FAT.  Required,
+ * for the Windows 98 FAT32 boot loader
+ */
 int
-fat_resize (PedFileSystem* fs, PedGeometry* geom)
+_copy_hidden_sectors (FatOpContext* ctx)
+{
+	FatSpecific*    old_fs_info = FAT_SPECIFIC (ctx->old_fs);
+	FatSpecific*    new_fs_info = FAT_SPECIFIC (ctx->new_fs);
+	PedSector       first = 1;
+	PedSector       last;
+	PedSector       count;
+
+	/* nothing to copy for FAT16 */
+	if (old_fs_info->fat_type == FAT_TYPE_FAT16
+			|| new_fs_info->fat_type == FAT_TYPE_FAT16)
+		return 1;
+
+	last = PED_MIN (old_fs_info->fat_offset, new_fs_info->fat_offset) - 1;
+	count = last - first + 1;
+
+	PED_ASSERT (count < BUFFER_SIZE, return 0);
+
+	if (!ped_geometry_read (ctx->old_fs->geom, old_fs_info->buffer,
+				first, count))
+		return 0;
+	if (!ped_geometry_write (ctx->new_fs->geom, old_fs_info->buffer,
+				 first, count))
+		return 0;
+	return 1;
+}
+
+int
+fat_resize (PedFileSystem* fs, const PedGeometry* geom)
 {
 	FatSpecific*	fs_info = FAT_SPECIFIC (fs);
 	FatSpecific*	new_fs_info;
@@ -694,16 +794,12 @@ fat_resize (PedFileSystem* fs, PedGeometry* geom)
 	if (!fat_table_write_all (new_fs_info->fat, new_fs))
 		goto error_abort_ctx;
 
-	if (!fat_boot_sector_generate (&new_fs_info->boot_sector, new_fs))
-		goto error_abort_ctx;
-	if (!fat_boot_sector_write (&new_fs_info->boot_sector, new_fs))
-		goto error_abort_ctx;
+	_copy_hidden_sectors (ctx);
+	fat_boot_sector_generate (&new_fs_info->boot_sector, new_fs);
+	fat_boot_sector_write (&new_fs_info->boot_sector, new_fs);
 	if (new_fs_info->fat_type == FAT_TYPE_FAT32) {
-		if (!fat_info_sector_generate (&new_fs_info->info_sector,
-					       new_fs))
-			goto error_abort_ctx;
-		if (!fat_info_sector_write (&new_fs_info->info_sector, new_fs))
-			goto error_abort_ctx;
+		fat_info_sector_generate (&new_fs_info->info_sector, new_fs);
+		fat_info_sector_write (&new_fs_info->info_sector, new_fs);
 	}
 
 	if (!resize_context_assimilate (ctx))
