@@ -28,6 +28,7 @@
 #include <parted/disk_pc98.h>
 #include <parted/disk_mac.h>
 #include <parted/disk_bsd.h>
+#include <parted/disk_gpt.h>
 #include <parted/disk_sun.h>
 
 #include <libintl.h>
@@ -38,7 +39,6 @@
 #endif /* ENABLE_NLS */
 
 #include <unistd.h>
-#include <asm/page.h>
 #include <string.h>
 
 #define SWAP_SPECIFIC(fs) ((SwapSpecific*) (fs->type_specific))
@@ -566,6 +566,15 @@ swap_set_system (const PedFileSystem* fs, PedPartition* part,
 		return 1;
 	}
 
+	if (strcmp (disk_type->name, GPT_NAME) == 0) {
+		GPTPartitionData* gpt_part_data = part->disk_specific;
+		PED_ASSERT(gpt_part_data      != NULL, return 0);
+		PED_ASSERT(gpt_part_data->pte != NULL, return 0);
+		gpt_part_data->pte->PartitionTypeGuid =
+			PARTITION_SWAP_GUID;
+		return 1;
+	}
+
 	if (strcmp (disk_type->name, SUN_NAME) == 0) {
 		SunPartitionData* sun_data = part->disk_specific;
 		sun_data->type = 0x82;
