@@ -26,13 +26,10 @@
 #include <parted/disk_mac.h>
 #include <parted/disk_pc98.h>
 #include <parted/disk_sun.h>
+#include <parted/disk_gpt.h>
 
-#include <grp.h>
-#include <pwd.h>
 #include <string.h>
 #include <stdlib.h>
-#include <unistd.h>
-#include <sys/types.h>
 
 #include <locale.h>
 #include <libintl.h>
@@ -45,28 +42,38 @@
 /* TODO: make file systems modules.  Actually this will probably require
  * 2 parts, because we always want clobber and probe to be available
  */
-extern void ped_file_system_ufs_init ();
-extern void ped_file_system_reiserfs_init ();
-extern void ped_file_system_ntfs_init ();
-extern void ped_file_system_linux_swap_init ();
-extern void ped_file_system_hfs_init ();
-extern void ped_file_system_fat_init ();
 extern void ped_file_system_ext2_init ();
+extern void ped_file_system_fat_init ();
+extern void ped_file_system_hfs_init ();
+extern void ped_file_system_jfs_init ();
+extern void ped_file_system_linux_swap_init ();
+extern void ped_file_system_ntfs_init ();
+extern void ped_file_system_reiserfs_init ();
+extern void ped_file_system_ufs_init ();
+extern void ped_file_system_xfs_init ();
 	
 extern void ped_file_system_ext2_done ();
 extern void ped_file_system_fat_done ();
 extern void ped_file_system_hfs_done ();
+extern void ped_file_system_jfs_done ();
 extern void ped_file_system_linux_swap_done ();
 extern void ped_file_system_ntfs_done ();
 extern void ped_file_system_reiserfs_done ();
 extern void ped_file_system_ufs_done ();
+extern void ped_file_system_xfs_done ();
 
 /* ped_malloc() debugging.  Stick the address and size of memory blocks that
  * weren't ped_free()d in here, and an exception will be thrown when it is
  * allocated.  That way, you can find out what, exactly, the allocated thing
  * is, and where it is created.
  */
-static int	dodgey_malloc_list[10][2] = {
+typedef struct
+{
+    void*	pointer;
+    size_t	size;
+} pointer_size_type;
+
+static pointer_size_type dodgey_malloc_list[] = {
  {0,		0},
  {0,		0},
  {0,		0},
@@ -89,19 +96,24 @@ init_disk_types ()
 {
 	ped_disk_loop_init ();
 	ped_disk_msdos_init ();
+#ifdef ENABLE_PC98
 	ped_disk_pc98_init ();
+#endif
 	ped_disk_mac_init ();
 	ped_disk_bsd_init ();
+	ped_disk_gpt_init ();
 	ped_disk_sun_init ();
 }
 
 static void
 init_file_system_types ()
 {
+	ped_file_system_xfs_init ();
 	ped_file_system_ufs_init ();
 	ped_file_system_reiserfs_init ();
 	ped_file_system_ntfs_init ();
 	ped_file_system_linux_swap_init ();
+	ped_file_system_jfs_init ();
 	ped_file_system_hfs_init ();
 	ped_file_system_fat_init ();
 	ped_file_system_ext2_init ();
@@ -110,9 +122,6 @@ init_file_system_types ()
 int
 ped_init ()
 {
-	if (getlogin())
-		initgroups (getlogin(), 0);
-
 #ifdef ENABLE_NLS
 	bindtextdomain (PACKAGE, LOCALEDIR);
 #endif
@@ -120,8 +129,6 @@ ped_init ()
 	init_disk_types ();
 	init_file_system_types ();
 
-	memset (dodgey_memory_active, 0, sizeof (dodgey_memory_active));
-
 	return 1;
 }
 
@@ -129,10 +136,13 @@ static void
 done_disk_types ()
 {
 	ped_disk_msdos_done ();
+#ifdef ENABLE_PC98
         ped_disk_pc98_done ();
+#endif
 	ped_disk_loop_done ();
 	ped_disk_mac_done ();
 	ped_disk_bsd_done ();
+	ped_disk_gpt_done ();
 	ped_disk_sun_done ();
 }
 
@@ -142,10 +152,12 @@ done_file_system_types ()
 	ped_file_system_ext2_done ();
 	ped_file_system_fat_done ();
 	ped_file_system_hfs_done ();
+	ped_file_system_jfs_done ();
 	ped_file_system_linux_swap_done ();
 	ped_file_system_ntfs_done ();
 	ped_file_system_reiserfs_done ();
 	ped_file_system_ufs_done ();
+	ped_file_system_xfs_done ();
 }
 
 void
@@ -168,10 +180,10 @@ _check_dodgey_pointer (const void* ptr, size_t size, int is_malloc)
 {
 	int		i;
 
-	for (i=0; dodgey_malloc_list[i][0]; i++) {
-		if (dodgey_malloc_list[i][0] != (int) ptr)
+	for (i=0; dodgey_malloc_list[i].pointer; i++) {
+		if (dodgey_malloc_list[i].pointer != ptr)
 			continue;
-		if (is_malloc && dodgey_malloc_list[i][1] != (int) size)
+		if (is_malloc && dodgey_malloc_list[i].size != size)
 			continue;
 		if (!is_malloc && !dodgey_memory_active[i])
 			continue;
