@@ -1,10 +1,6 @@
 /*
     libparted - a library for manipulating disk partitions
-    Copyright (C) 1998-2000  Andrew Clausen, Lennert Buytenhek and Red Hat Inc.
-
-	Andrew Clausen			<clausen@gnu.org>
-	Lennert Buytenhek		<buytenh@gnu.org>
-	Matt Wilson, Red Hat Inc.	<msw@redhat.com>
+    Copyright (C) 2000 Free Software Foundation, Inc.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
