@@ -84,22 +84,6 @@ static char* flag_msg;
 
 static Command*	commands [256] = {NULL};
 
-static int do_boot (PedDevice** dev);
-static int do_check (PedDevice** dev);
-static int do_cp (PedDevice** dev);
-int do_help (PedDevice** dev);
-static int do_hide (PedDevice** dev);
-static int do_mklabel (PedDevice** dev);
-static int do_mkfs (PedDevice** dev);
-static int do_mkpart (PedDevice** dev);
-static int do_mkpartfs (PedDevice** dev);
-static int do_print (PedDevice** dev);
-static int do_quit (PedDevice** dev);
-static int do_resize (PedDevice** dev);
-static int do_rm (PedDevice** dev);
-static int do_unhide (PedDevice** dev);
-static int do_name (PedDevice** dev);
-
 static void _done (PedDevice* dev);
 
 void
@@ -113,6 +97,45 @@ help_on (char* topic)
 	command_print_help (cmd);
 }
 
+static inline PedSector
+_ped_abs (PedSector sector)
+{
+	return sector > 0 ? sector : -sector;
+}
+
+/* check if what the user will get is roughly what the user wanted.  If it isn't
+ * then ask them if they like what they're going to get.  (That's what msg is
+ * for).   msg should contain 4 %f's... requested start/end, and the solution
+ * start/end.
+ * 	Returns 1 iff all is ok.
+ */
+static int
+_solution_check_distant (PedSector req_start, PedSector req_end,
+			 PedSector soln_start, PedSector soln_end,
+			 const char* msg)
+{
+	PedSector	start_delta = _ped_abs (soln_start - req_start);
+	PedSector	end_delta = _ped_abs (soln_end - req_end);
+	PedSector	distance = start_delta + end_delta;
+
+	if (distance == -1
+	    || (distance > 10 * MEGABYTE_SECTORS
+		    && 1.0 * distance / (req_end - req_start + 1) > 1.1)) {
+		if (ped_exception_throw (
+			PED_EXCEPTION_WARNING,
+			PED_EXCEPTION_OK_CANCEL,
+			msg,
+			req_start * 1.0 / MEGABYTE_SECTORS,
+			req_end * 1.0 / MEGABYTE_SECTORS,
+			soln_start * 1.0 / MEGABYTE_SECTORS,
+			soln_end * 1.0 / MEGABYTE_SECTORS)
+				== PED_EXCEPTION_CANCEL)
+			return 0;
+	}
+
+	return 1;
+}
+
 static int
 do_check (PedDevice** dev)
 {
@@ -440,6 +463,12 @@ do_mkpart (PedDevice** dev)
 	if (!ped_disk_add_partition (disk, part, constraint))
 		goto error_destroy_part;
 
+	if (!_solution_check_distant (start, end,
+				      part->geom.start, part->geom.end,
+		_("You requested to create a partition at %.3f-%.3fMb. The "
+		  "closest Parted can manage is %.3f-%.3fMb.")))
+		goto error_remove_part;
+
 	/* FIXME:  ped_partition_new() doesn't probe the file system for
 	 * what type of partition should be created.  So we do it here.
 	 * 	Question: SHOULD we probe the file system on
@@ -453,6 +482,8 @@ do_mkpart (PedDevice** dev)
 	ped_disk_close (disk);
 	return 1;
 
+error_remove_part:
+	ped_disk_remove_partition (disk, part);
 error_destroy_part:
 	ped_partition_destroy (part);
 error_destroy_constraint:
@@ -526,6 +557,12 @@ do_mkpartfs (PedDevice** dev)
 	if (!ped_disk_add_partition (disk, part, constraint))
 		goto error_destroy_part;
 
+	if (!_solution_check_distant (start, end,
+				      part->geom.start, part->geom.end,
+		_("You requested to create a partition at %.3f-%.3fMb. The "
+		  "closest Parted can manage is %.3f-%.3fMb.")))
+		goto error_remove_part;
+
 	fs = ped_file_system_create (&part->geom, fs_type);
 	if (!fs) 
 		goto error_remove_part;
@@ -542,8 +579,7 @@ do_mkpartfs (PedDevice** dev)
 	return 1;
 
 error_remove_part:
-	ped_disk_delete_partition (disk, part);
-	goto error_close_disk;
+	ped_disk_remove_partition (disk, part);
 error_destroy_part:
 	ped_partition_destroy (part);
 error_destroy_constraint:
@@ -633,6 +669,12 @@ do_move (PedDevice** dev)
 		goto error_close_disk;
 	}
 
+	if (!_solution_check_distant (start, end,
+				      part->geom.start, part->geom.end,
+		_("You requested to move the partition to %.3f-%.3fMb. The "
+		  "closest Parted can manage is %.3f-%.3fMb.")))
+		goto error_close_disk;
+
 /* do the move */
 	fs = ped_file_system_open (&old_geom);
 	if (!fs)
@@ -837,6 +879,12 @@ do_resize (PedDevice** dev)
 		if (!ped_disk_set_partition_geom (disk, part, constraint,
 						  start, end))
 			goto error_close_fs;
+		if (!_solution_check_distant (start, end,
+					      part->geom.start, part->geom.end,
+			_("You requested to resize the partition to "
+			  "%.3f-%.3fMb. The closest Parted can manage is "
+			  "%.3f-%.3fMb.")))
+			goto error_close_fs;
 		if (!ped_file_system_resize (fs, &part->geom))
 			goto error_close_fs;
 		ped_file_system_close (fs);
@@ -1304,9 +1352,9 @@ main (int argc, char** argv)
 		return 0;
 
 	if (argc)
-		status = non_interactive_mode (dev, commands, argc, argv);
+		status = non_interactive_mode (&dev, commands, argc, argv);
 	else
-		status = interactive_mode (dev, commands);
+		status = interactive_mode (&dev, commands);
 
 	_done (dev);
 
