@@ -19,6 +19,12 @@
 
 #include <parted/parted.h>
 
+#include <ctype.h>
+#include <signal.h>
+#include <stdlib.h>
+#include <string.h>
+#include <unistd.h>
+
 #include "../config.h"
 #include "command.h"
 #include "strlist.h"
@@ -54,12 +60,6 @@ extern int tgetnum (char* key);
 
 #endif /* HAVE_LIBREADLINE */
 
-#include <ctype.h>
-#include <signal.h>
-#include <stdlib.h>
-#include <string.h>
-#include <unistd.h>
-
 char* prog_name = "GNU Parted " VERSION "\n";
 
 static char* banner_msg = N_(
@@ -529,45 +529,106 @@ help_msg ()
 	exit (0);
 }
 
+/* we need a CHS warning if there is a DOS partition table with a FAT
+ * partition(s) without the LBA flag set
+ */
+static int
+_need_chs_warning (PedDevice* dev)
+{
+#if defined (__i386__)
+	PedDisk*	disk;
+	PedPartition*	walk;
+	int		has_non_lba_fat;
+
+	disk = ped_disk_open (dev);
+	if (!disk)
+		return 0;
+
+	if (strcmp (disk->type->name, "msdos")) {
+		ped_disk_close (disk);
+		return 0;
+	}
+
+	has_non_lba_fat = 0;
+	for (walk = ped_disk_next_partition (disk, NULL); walk;
+	     walk = ped_disk_next_partition (disk, walk)) {
+		if (!ped_partition_is_active (walk))
+			continue;
+		if (!walk->fs_type)
+			continue;
+
+		if (!strcmp(walk->fs_type->name, "FAT")
+		    && !ped_partition_get_flag (walk, PED_PARTITION_LBA))
+			has_non_lba_fat = 1;
+	}
+
+	ped_disk_close (disk);
+	return has_non_lba_fat;
+#else
+	return 0;
+#endif
+}
+
+/* if we need to scare users about CHS stuff, do it...
+ */
+static void
+_do_chs_check (PedDevice* dev)
+{
+#if defined (__i386__)
+	PedExceptionType	ex_type;
+	char*			warning_msg;
+
+	if (_need_chs_warning (dev)) {
+		ex_type = PED_EXCEPTION_WARNING;
+		warning_msg = _("  You have Windows FAT partition(s) that are "
+				"not using LBA.  If your BIOS supports LBA, "
+				"then you should switch to LBA by setting the "
+				"LBA flag on all FAT partitions.  Otherwise, "
+				"make sure the operating system and the BIOS "
+				"have the same geometry before resizing any "
+				"FAT partitions.");
+	} else {
+		ex_type = PED_EXCEPTION_INFORMATION;
+		warning_msg = "";
+	}
+
+	/* already been subjected to a rant */
+	if (!dev->geom_known)
+		return;
+
+	if (dev->cylinders > 1024)
+		ped_exception_throw (
+			ex_type,
+			PED_EXCEPTION_OK,
+		_("The operating system thinks the geometry on %s is "
+		  "%d/%d/%d.  Therefore, cylinder 1024 ends at %.3fM.%s"),
+			dev->path,
+			dev->cylinders, dev->heads, dev->sectors,
+			(1024 * dev->heads * dev->sectors - 1) / 2 / 1024.0,
+			warning_msg);
+	else
+		ped_exception_throw (
+			ex_type,
+			PED_EXCEPTION_OK,
+		_("The operating system thinks the geometry on %s is "
+		  "%d/%d/%d.%s"),
+			dev->path,
+			dev->cylinders, dev->heads, dev->sectors,
+			warning_msg);
+#endif
+}
+
 void
 print_using_dev (PedDevice* dev)
 {
 	printf (_("\nUsing %s\n"), dev->path);
-
-#if defined (__i386__)
 	if (dev->type != PED_DEVICE_UNKNOWN
-	    && dev->type != PED_DEVICE_FILE) {
-		if (dev->cylinders > 1024)
-			ped_exception_throw (
-				PED_EXCEPTION_WARNING,
-				PED_EXCEPTION_OK,
-			_("The operating system thinks the geometry on %s is "
-			  "%d/%d/%d.  Therefore, cylinder 1024 ends at %.3fM.  "
-			  "You should check that this matches the BIOS "
-			  "geometry before using this program."),
-				dev->path,
-				dev->cylinders,
-				dev->heads,
-				dev->sectors,
-				(1024 * dev->heads * dev->sectors - 1)
-					/ 2 / 1024.0);
-		else
-			ped_exception_throw (
-				PED_EXCEPTION_WARNING,
-				PED_EXCEPTION_OK,
-			_("The operating system thinks the geometry on %s is "
-			  "%d/%d/%d.  You should check that this matches the "
-			  "BIOS geometry before using this program."),
-				dev->path,
-				dev->cylinders,
-				dev->heads,
-				dev->sectors);
-	}
-#endif
+	    && dev->type != PED_DEVICE_FILE)
+		_do_chs_check (dev);
 }
 
 int
-interactive_mode (PedDevice* dev, Command* cmd_list[])
+interactive_mode (PedDevice** dev, Command* cmd_list[])
 {
 	char*		line;
 	Command*	cmd;
@@ -581,7 +642,7 @@ interactive_mode (PedDevice* dev, Command* cmd_list[])
 	str_list_print_wrap (list, screen_width (), 0, 0);
 	str_list_destroy (list);
 
-	print_using_dev (dev);
+	print_using_dev (*dev);
 
 	while (1) {
 		line = _readline ("(parted) ");
@@ -595,7 +656,7 @@ interactive_mode (PedDevice* dev, Command* cmd_list[])
 
 		cmd = command_get (commands, get_word ());
 		if (cmd)
-			command_run (cmd, &dev);
+			command_run (cmd, dev);
 		else
 			print_commands_help ();
 
@@ -607,7 +668,7 @@ interactive_mode (PedDevice* dev, Command* cmd_list[])
 }
 
 int
-non_interactive_mode (PedDevice* dev, Command* cmd_list[],
+non_interactive_mode (PedDevice** dev, Command* cmd_list[],
 		      int argc, char* argv[])
 {
 	int		i;
@@ -626,7 +687,7 @@ non_interactive_mode (PedDevice* dev, Command* cmd_list[],
 			help_msg ();
 			goto error;
 		}
-		if (!command_run (cmd, &dev))
+		if (!command_run (cmd, dev))
 			goto error;
 	}
 
